#include <assert.h>
#include <string.h>
#include "colors.h"
#include "combat.h"
#include "map.h"
#include "messages.h"
#include "misc.h"
#include "monster.h"
#include "player.h"
#include "vision.h"

static int monprob;

static void addislot(monster *mn, int part);

monster *createmonster(int type, coord where)
{ monster *mn;

  mn=zmalloc(sizeof(monster));
  mn->type=type;
  mn->where=where;
  mn->vision=monlist[type].vision;
  mn->speed=monlist[type].speed;
  mn->hp=mn->maxhp=roll(monlist[type].hp);
  mn->strength=monlist[type].strength;
  mn->dexterity=mn->maxdexterity=monlist[type].dexterity;
  mn->intelligence=monlist[type].intelligence;
  mn->protection=monlist[type].intrinsic_pv;
  mn->healspeed=100;

  switch(monlist[type].class)
  { case HUMANOID:
      addislot(mn, S_FINGER);
      addislot(mn, S_FINGER);
      addislot(mn, S_HAND);
      addislot(mn, S_HAND);
      addislot(mn, S_FEET);
      addislot(mn, S_GLOVES);
      addislot(mn, S_BODY);
      addislot(mn, S_HEAD);
      addislot(mn, S_NECK);
      break;
  }

  return mn;
}

void removemonster(monster *mn)
{ mn->prev->next=mn->next;
  mn->next->prev=mn->prev;
}

static void addislot(monster *mn, int part)
{ islot *b;

  b=zmalloc(sizeof(islot));
  b->type=part;
  b->next=mn->parts;
  mn->parts=b;
}

static int followscent(monster *mn)
{ int i, j, x=mn->where.x, y=mn->where.y;

  for(i=-1; i<=1; i++)
    for(j=-1; j<=1; j++)
    { if(can_go((coord){ mn->where.x+i, mn->where.y+j }) &&
         m->place[mn->where.x+i][mn->where.y+j].scent>m->place[x][y].scent)
        x=mn->where.x+i, y=mn->where.y+j;
    }

  if(mn->where.x==x && mn->where.y==y) return 0;
  if(can_go((coord) { x, y })) mn->where.x=x, mn->where.y=y;

  return 1;
}

static void movetowardsplayer(monster *mn)
{ coord nw;
  int xv, yv;

  nw=mn->where;
  xv=sgn(you.where.x-nw.x);
  yv=sgn(you.where.y-nw.y);

  if(can_go((coord){ nw.x+xv, nw.y+yv }))
  { mn->where.x+=xv;
    mn->where.y+=yv;
  }
  else if(can_go((coord){ nw.x, nw.y+yv })) mn->where.y+=yv;
  else if(can_go((coord){ nw.x+xv, nw.y })) mn->where.x+=xv;
}

static void moverandomly(monster *mn)
{ int i;

  for(i=0; i<8; i++)
  { coord w;
    w.x=mn->where.x+randint(3)-1;
    w.y=mn->where.y+randint(3)-1;

    if(can_go(w))
    { mn->where=w;
      return;
    }
  }
}

static void movemonster(monster *mn)
{ if(distance(mn->where, you.where)==1) attack(mn, &you);
  else if(testline(mn->where, you.where)) movetowardsplayer(mn);
  else if(!followscent(mn)) moverandomly(mn);
}

void movemonsters()
{ monster *mn;

  updatedisplay();

  for(mn=m->mhead->next; mn; mn=mn->next)
  { if(!mn->movetimeout--)
    { mn->movetimeout=mn->speed;
      movemonster(mn);
    }
    if(!mn->healtimeout--)
    { mn->healtimeout=mn->healspeed;
      if(mn->hp<mn->maxhp) mn->hp++;
    }
  }
}

monster *occupied(coord w)
{ monster *mn;

  for(mn=m->mhead->next; mn; mn=mn->next)
    if(mn->where.x==w.x && mn->where.y==w.y) return mn;

  if(you.where.x==w.x && you.where.y==w.y) return &you;

  return 0;
}

monster *addmonster(monster *mn)
{ monster *n;

  for(n=m->mhead; n->next; n=n->next);
  n->next=mn;
  mn->prev=n;

  return mn;
}

int can_go(coord w)
{ return walkable(w) && !occupied(w);
}

char *slotname(int t)
{ switch(t)
  { case S_HEAD: return "head";
    case S_FEET: return "feet";
    case S_HAND: return "hand";
    case S_GLOVES: return "pair of hands";
    case S_BODY: return "body";
    case S_FINGER: return "ring finger";
    case S_NECK: return "neck";
  }

  return 0;
}

char *Slotname(int t)
{ switch(t)
  { case S_HEAD: return "Head";
    case S_FEET: return "Feet";
    case S_HAND: return "Hand";
    case S_GLOVES: return "Gloves";
    case S_BODY: return "Body";
    case S_FINGER: return "Finger";
    case S_NECK: return "Neck";
  }

  return 0;
}

char *monname(monster *mn)
{ static char buf[BUFSZ];

  if(mn==&you) return "you";

  strcpy(buf, "the ");
  strcat(buf, monlist[mn->type].name);

  return buf;
}

char *Monname(monster *mn)
{ static char buf[BUFSZ];

  if(mn==&you) return "You";

  strcpy(buf, "The ");
  strcat(buf, monlist[mn->type].name);

  return buf;
}

void initmonprobs()
{ int i;

  for(i=0; i<N_MONS; i++)
    monprob+=monlist[i].probability;
}

int randmonster(int level)
{ int i, n;

  do
  { n=randint(monprob);
    for(i=0; i<N_MONS && n>=0; i++) n-=monlist[i].probability;
    i--;
    assert(i<N_MONS);
  } while(monlist[i].difficulty>level);

  return i;
}
