#include <assert.h>
#include <string.h>
#include "colors.h"
#include "map.h"
#include "messages.h"
#include "misc.h"
#include "monster.h"
#include "objects.h"
#include "objlist.h"
#include "player.h"

static int objprob;

char *classnames[]=
{ "Weapons",
  "Armor",
  "Food"
};

obj *createobject(int type)
{ obj *o;

  o=zmalloc(sizeof(obj));
  o->type=type;

  return o;
}

obj *addobjtotile(coord c, obj *o)
{ o->prev=0;
  if(m->place[c.x][c.y].items) m->place[c.x][c.y].items->prev=o;
  o->next=m->place[c.x][c.y].items;
  m->place[c.x][c.y].items=o;

  return o;
}

obj *addobjtomonster(monster *mn, obj *o)
{ char c;

  if(mn==&you)
  { c=findfreeletter(mn->inventory, o->letter);
    if(!c) return 0;
    o->letter=c;
  }

  o->prev=0;
  if(mn->inventory) mn->inventory->prev=o;
  o->next=mn->inventory;
  mn->inventory=o;

  return o;
}

obj *removeobjfromtile(coord c, obj *o)
{ if(m->place[c.x][c.y].items==o) m->place[c.x][c.y].items=o->next;
  else if(o->prev) o->prev->next=o->next;
  if(o->next) o->next->prev=o->prev;

  o->next=o->prev=0;

  return o;  
}

obj *removeobjfrommonster(monster *m, obj *o)
{ if(m->inventory==o) m->inventory=o->next;
  else if(o->prev) o->prev->next=o->next;
  if(o->next) o->next->prev=o->prev;

  o->next=o->prev=0;

  return o;
}

char *objname(obj *o)
{ static char buf[BUFSZ];

  if(!o) return strcpy(buf, "nothing");
  switch(objlist[o->type].article)
  { case ARTICLE_A: strcpy(buf, "a "); break;
    case ARTICLE_AN: strcpy(buf, "an "); break;
    case ARTICLE_SOME: strcpy(buf, "some "); break;
    case ARTICLE_PAIR: strcpy(buf, "a pair of "); break;
    default: *buf=0; break;
  }
  strcat(buf, objlist[o->type].name);

  return buf;
}

char *the_objname(obj *o)
{ static char buf[BUFSZ];

  if(!o) return strcpy(buf, "nothing");
  strcpy(buf, "the ");
  strcat(buf, objlist[o->type].name);

  return buf;
}

char *Objname(obj *o)
{ static char buf[BUFSZ];

  if(!o) return strcpy(buf, "Nothing");
  switch(objlist[o->type].article)
  { case ARTICLE_A: strcpy(buf, "A "); break;
    case ARTICLE_AN: strcpy(buf, "An "); break;
    case ARTICLE_SOME: strcpy(buf, "Some "); break;
    case ARTICLE_PAIR: strcpy(buf, "A pair of "); break;
    default: *buf=0; break;
  }
  strcat(buf, objlist[o->type].name);

  return buf;
}

void listcontents(obj *o)
{ while(o)
  { if(o->next) pline("%s,", objname(o));
    else if(o->prev) pline("and %s.", objname(o));
    else pline("%s.", objname(o));
    o=o->next;
  }
}

void initobjprobs()
{ int i;

  for(i=0; i<N_OBJS; i++)
    objprob+=objlist[i].probability;
}

int randobject()
{ int i, n;

  n=randint(objprob);
  for(i=0; i<N_OBJS && n>=0; i++) n-=objlist[i].probability;
  i--;
  assert(i<N_OBJS);

  return i;
}
