#include <ctype.h>
#include <stdlib.h>
#include "combat.h"
#include "io.h"
#include "map.h"
#include "monster.h"
#include "player.h"
#include "messages.h"
#include "objects.h"
#include "ticks.h"
#include "vision.h"

monster *player;
extrastuff pstuff;

int telepathy;

#define ONETURN 1

static coord movedir(coord w, char dir);
static int isdir(char dir);
static char handlerun(char dir);

static int doclose();
static int dodrop();
static int doinventory();
static int dokick();
static int dolook();
static int doopen();
static int dopickup();
static int doviewworn();
static int doequip();
static int dodeequip();
static int dotelepathy();
static int dopastmessages();
static int dogoup();
static int dogodown();
static int dorun();
static int doeat();

static void showcontents(coord w);

struct
{ char c;
  int (*func)();
} actlist[]=
{ { ',', dopickup },
  { ':', dolook },
  { '<', dogoup },
  { '>', dogodown },
  { 'c', doclose },
  { 'd', dodrop },
  { 'e', doeat },
  { 'i', doinventory },
  { 'k', dokick },
  { 'm', dopastmessages },
  { 'o', doopen },
  { 'p', dotelepathy },
  { 'r', dorun },
  { 't', dodeequip },
  { 'v', doviewworn },
  { 'w', doequip }
};

int playerinput()
{ char c;
  coord w=you.where;
  int i;

  if(you.hp<1) return 0;

  for(;;)
  { updatedisplay();
    cleartopline();
    placecursor(you.where.x, you.where.y);

    if(pstuff.running) c=handlerun(pstuff.direction);
    else c=readchar();

    if(c=='.' || c=='5') return 1;

    if(isdir(c))
    { coord new=movedir(w, c-'0');
      if(!walkable(new)) You("can't go that way.");
      else if(occupied(new))
      { attack(player, occupied(new));
        return 1;
      }
      else
      { you.where=new;
        if(m->place[new.x][new.y].items) showcontents(new);
        return 1;
      }
    }  

    if(c=='q' && tolower(prompt("Really quit [yN]?"))=='y') return 0;

    for(i=0; actlist[i].c; i++)
      if(actlist[i].c==c && actlist[i].func()) return pstuff.level;
  }

  return 1;
}

char findfreeletter(obj *o, char opt)
{ char lcase[26], ucase[26];
  int i;

  for(i=0; i<26; i++) lcase[i]=ucase[i]=0;

  while(o)
  { if(o->letter>='a' && o->letter<='z') lcase[o->letter-'a']=1;
    else ucase[o->letter-'A']=1;
    o=o->next;
  }

  if(opt && ((opt>='a' && opt<='z')?lcase:ucase)[(int)opt]) return opt;
  for(i=0; i<26; i++) if(!lcase[i]) return i+'a';
  for(i=0; i<26; i++) if(!ucase[i]) return i+'A';

  return 0;
}

static char getdir()
{ char d;

  d=prompt("Which direction?");
  if(!isdir(d)) return 0;

  return d;
}

static obj *getitem(char c)
{ obj *o;

  for(o=you.inventory; o && o->letter!=c; o=o->next);
  if(!o) pline("You're not carrying that.");
  return o;
}

static int doclose()
{ char d;
  coord w;

  if(!(d=getdir())) return 0;
  w=movedir(you.where, d-'0');

  if(here(w.x, w.y)==BROKEN_DOOR)
  { pline("That door is broken.");
    return 0;
  }

  if(here(w.x, w.y)!=DOOR)
  { You("see no door there.");
    return 0;
  }

  if(special(w.x, w.y).closed)
  { pline("That door is already closed!");
    return 0;
  }

  if(occupied(w) || m->place[w.x][w.y].items)
  { pline("Something's in the way.");
    return 0;
  }

  pline("The door closes.");
  special(w.x, w.y).closed=CLOSED;

  return ONETURN;
}

static int dodrop()
{ obj *o;

  if(!(o=getitem(prompt("What would you like to drop?")))) return 0;

  if(o->worn)
  { You("can't drop something you have equipped.");
    return 0;
  }

  addobjtotile(you.where, removeobjfrommonster(&you, o));
  pline("Dropped: %s.", objname(o));

  return ONETURN;
}

static int dokick()
{ char d;
  coord w;

  if(!(d=getdir())) return 0;
  w=movedir(you.where, d-'0');

  if(here(w.x, w.y)==WALL) You("kick the wall.");
  else
    if(here(w.x, w.y)==DOOR)
    { pline("The door crashes open!");
      here(w.x, w.y)=BROKEN_DOOR;
    }
  else You("kick at empty space.");
  
  return ONETURN;
}

int lcase(char c)
{ return c>='a' && c<='z';
}

/* sort by object class, then by case, then by letter */
static int inv_cmp(const void *c1, const void *c2)
{ obj *d1=*(obj **)c1, *d2=*(obj **)c2;

  if(objlist[d1->type].class==objlist[d2->type].class)
  { if(lcase(d1->letter)!=lcase(d2->letter)) return d1->letter<d2->letter;
    else return d1->letter>d2->letter;
  }
  else return objlist[d1->type].class>objlist[d2->type].class;
}

static int doinventory()
{ obj *o, *list[52];
  int i, n=0, cclass=-1;

  if(!you.inventory)
  { pline("You are empty-handed.");
    return 0;
  }

  for(o=you.inventory; o; o=o->next) list[n++]=o;

  qsort((void *)list, n, sizeof(obj *), inv_cmp);

  print_list("You're holding:");
  for(i=0; i<n; i++)
  { if(cclass!=objlist[list[i]->type].class)
      print_list(classnames[cclass=objlist[list[i]->type].class]);
    print_list(" %c - %s%s", list[i]->letter, objname(list[i]),
                             list[i]->worn?" (equipped)":"");
  }
  end_list();

  return 0;
}

static int dolook()
{ obj *o=m->place[you.where.x][you.where.y].items;

  if(!o) You("see no objects here.");
  else if(!o->next) showcontents(you.where);
  else
  { You("see here");
    listcontents(o);
  }

  return 0;
}

static int doopen()
{ char d;
  coord w;

  if(!(d=getdir())) return 0;
  w=movedir(you.where, d-'0');

  if(here(w.x, w.y)==BROKEN_DOOR)
  { pline("That door is broken.");
    return 0;
  }

  if(here(w.x, w.y)!=DOOR)
  { You("see no door there.");
    return 0;
  }

  if(!special(w.x, w.y).closed)
  { pline("That door is already open!");
    return 0;
  }

  if(special(w.x, w.y).closed==LOCKED)
    pline("That door seems to be locked.");
  else
  { pline("The door opens.");
    special(w.x, w.y).closed=OPEN;
  }

  return ONETURN;
}

static int dopickup()
{ obj *o=m->place[you.where.x][you.where.y].items;

  if(!o)
  { pline("There is nothing to pick up!");
    return 0;
  }

  if(o->next)
  { for(; o; o=o->next)
    { char c;
      c=prompt("Pick up %s [Ynq]?", the_objname(o));
      if(c=='q' || c==27) return 0;
      if(c!='n' && c!='N') break;
    }
    if(!o) return 0;
  }

  if(!addobjtomonster(&you, removeobjfromtile(you.where, o)))
  { pline("Your knapsack cannot accomodate any more items.");
    return 0;
  }

  pline("%c - %s.", you.inventory->letter, objname(you.inventory));
  return ONETURN;
}

static int doviewworn()
{ islot *b;

  print_list("Equipment:");
  for(b=you.parts; b; b=b->next)
    print_list("%-7s: %s", Slotname(b->type),
                            b->wearing?objname(b->wearing):"empty");
  end_list();

  return 0;
}

static int doequip()
{ obj *o;
  islot *b;

  if(!(o=getitem(prompt("What would you like to equip?")))) return 0;

  if(o->worn)
  { You("already have that equipped.");
    return 0;
  }

  if(!objlist[o->type].slot)
  { You("can't equip that.");
    return 0;
  }

  for(b=you.parts; b && !(b->type==objlist[o->type].slot && !b->wearing);
      b=b->next);

  if(!b)
  { You("would need another free %s to equip that.",
        slotname(objlist[o->type].slot));
    return 0;
  }

  b->wearing=o;
  o->worn=1;
  you.protection+=objlist[o->type].n0;
  you.dexterity+=objlist[o->type].n1;

  pline("You equip %s.", the_objname(o));

  return ONETURN;
}

static int dodeequip()
{ obj *o;
  islot *b;

  if(!(o=getitem(prompt("What would you like to deequip?")))) return 0;

  if(!o->worn)
  { You("don't have that equipped.");
    return 0;
  }

  for(b=you.parts; b->wearing!=o; b=b->next);
  b->wearing=0;
  o->worn=0;
  you.protection-=objlist[o->type].n0;
  you.dexterity-=objlist[o->type].n1;

  You("deequip %s.", the_objname(o));

  return ONETURN;
}

static int dotelepathy()
{ telepathy=!telepathy;
  pline("Telepathy %s", telepathy?"on":"off");

  return 0;
}

static int dopastmessages()
{ showpastmessages();

  return 0;
}

static int dogodown()
{ if(here(you.where.x, you.where.y)!=STAIRWAY_DOWN)
  { You("can't go down here.");
    return 0;
  }

  changelevel(special(you.where.x, you.where.y).destination,
              pstuff.level, pstuff.level+2, STAIRWAY_UP);

  return ONETURN;
}

static int dogoup()
{ if(here(you.where.x, you.where.y)!=STAIRWAY_UP)
  { You("can't go up here.");
    return 0;
  }

  if(pstuff.level>1)
    changelevel(special(you.where.x, you.where.y).destination,
                pstuff.level-2, pstuff.level, STAIRWAY_DOWN);
  else
  { char c=prompt("There will be no return!  Still climb [yN]?");
    if(c=='y' || c=='Y') pstuff.level--;
  }

  return ONETURN;
}

static int dorun()
{ if(!(pstuff.direction=getdir())) return 0;
  pstuff.running=1;

  return 0;
}

static int doeat()
{ obj *o;

  if(!(o=getitem(prompt("What would you like to eat?")))) return 0;

  switch(objlist[o->type].material)
  { case ITALIAN:
      pline("That's italian food!!");
    case FLESH:
    case VEGETABLE:
      You("begin to eat %s.", the_objname(o));
      if(!gameticks(objlist[o->type].weight)) return 0;
      You("finish eating %s.", the_objname(o));
      if(objlist[o->type].material==ITALIAN) You("feel corrupted.");
      zfree(removeobjfrommonster(&you, o));
      break;
    default:
      pline("That's plainly inedible.");
  }

  return 0;
}

static coord movedir(coord w, char dir)
{ switch(dir)
  { case 7: w.y--; w.x--; break;
    case 4: w.x--; break;
    case 1: w.y++; w.x--; break;
    case 2: w.y++; break;
    case 3: w.x++; w.y++; break;
    case 6: w.x++; break;
    case 9: w.x++; w.y--; break;
    case 8: w.y--; break;
  }

  return w;
}

static int isdir(char dir)
{ return (dir>'0' && dir<='9') || dir=='.';
}

static void showcontents(coord w)
{ You("see here");
  if(m->place[w.x][w.y].items->next) pline("several objects.");
  else listcontents(m->place[w.x][w.y].items);
}

void createplayer()
{ player=createmonster(HUMAN, (coord) { 0, 0 });
  you.healspeed=200;
  addobjtomonster(&you, createobject(SHORT_SWORD));
  addobjtomonster(&you, createobject(MEATBALL));
  addobjtomonster(&you, createobject(SPAGHETTI));
  addobjtomonster(&you, createobject(PARTY_HAT));
  addobjtomonster(&you, createobject(TUXEDO));
  addobjtomonster(&you, createobject(POLKA_DOTTED_NECKTIE));
  addobjtomonster(&you, createobject(RUBBER_GLOVES));
  addobjtomonster(&you, createobject(ARCTIC_CIRCLE_CANDY));
}

char handlerun(char d)
{ coord c=movedir(you.where, d-'0');
  if(!can_go(c))
  { pstuff.running=0;
    return 0;
  }

  return d;
}
