#include "colors.h"
#include "io.h"
#include "map.h"
#include "messages.h"
#include "monster.h"
#include "objects.h"
#include "player.h"
#include "visinfo.h"
#include "vision.h"

static void castline(coord start, coord *line, int dist);
char visible[SCREEN_WIDTH][SCREEN_HEIGHT];

struct
{ int type;
  char chr;
  char color;
} dispinfo[]=
{ { FLOOR, '.', LIGHTGREY },
  { WALL, '#', DARKGREY },
  { BROKEN_DOOR, '.', LIGHTGREY },
  { STAIRWAY_UP, '<', LIGHTGREY },
  { STAIRWAY_DOWN, '>', LIGHTGREY },
  { 0, 0, 0 }
};

void updatedisplay()
{ int i, j, k;
  monster *mn;

  for(i=0; i<m->xsize; i++)
    for(j=0; j<m->ysize; j++)
      visible[i][j]=0;

  visible[you.where.x][you.where.y]=1;

  for(i=0; i<13; i++) castline(you.where, vislines[i], you.vision);

  for(i=1; i<SCREEN_WIDTH-1; i++)
    for(j=1; j<SCREEN_HEIGHT-1; j++)
    { screen[i][j][0]=m->place[i][j].memory;
      screen[i][j][1]=DARKGREY;

      if(visible[i][j])
      { if(m->place[i][j].items)
        { int type=m->place[i][j].items->type;
          screen[i][j][0]=objlist[type].chr;
          screen[i][j][1]=objlist[type].color;
        }
        else switch(here(i, j))
        { case DOOR:
            screen[i][j][0]=special(i, j).closed?'+':'\'';
            screen[i][j][1]=BROWN;
            break;
          default:
            for(k=0; dispinfo[k].color; k++)
              if(dispinfo[k].type==here(i, j))
              { screen[i][j][0]=dispinfo[k].chr;
                screen[i][j][1]=dispinfo[k].color;
                break;
              }
            break;
        }
      }
    }

  for(i=0; i<SCREEN_WIDTH; i++)
    for(j=0; j<SCREEN_HEIGHT; j++)
      m->place[i][j].memory=screen[i][j][0];

  for(mn=m->mhead->next; mn; mn=mn->next)
    if(visible[mn->where.x][mn->where.y] || telepathy)
    { screen[mn->where.x][mn->where.y][0]=monlist[mn->type].chr;
      screen[mn->where.x][mn->where.y][1]=monlist[mn->type].color;
    }

  screen[you.where.x][you.where.y][0]=monlist[you.type].chr;
  screen[you.where.x][you.where.y][1]=monlist[you.type].color;

  bottomline();
  updatescreen();
}

static int viewable(coord s)
{ if(here(s.x, s.y)==WALL) return 0;
  if(here(s.x, s.y)==DOOR && special(s.x, s.y).closed) return 0;
  return 1;
}

#define VIS_TEST(n, op1, op2, flip)                                  \
  (!stopped[n] &&                                                    \
   (visible[s.x op1 ((flip)?(line[i].x):(line[i].y))]                \
           [s.y op2 ((flip)?(line[i].y):(line[i].x))]=1) &&          \
   (line[i].x*line[i].x+line[i].y*line[i].y>dist*dist+1 ||           \
    !viewable((coord) { s.x op1 ((flip)?(line[i].x):(line[i].y)),    \
                        s.y op2 ((flip)?(line[i].y):(line[i].x))     \
                      }                                              \
             )                                                       \
   ) &&                                                              \
   (stopped[n]=1)                                                    \
  )


static void castline(coord s, coord *line, int dist)
{ int i;
  char stopped[8]={ 0, 0, 0, 0, 0, 0, 0, 0 };

  for(i=0; i<VISRANGE; i++)
  { VIS_TEST(0, +, +, 0);
    VIS_TEST(1, +, -, 0);
    VIS_TEST(2, -, +, 0);
    VIS_TEST(3, -, -, 0);
    VIS_TEST(4, +, +, 1);
    VIS_TEST(5, +, -, 1);
    VIS_TEST(6, -, +, 1);
    VIS_TEST(7, -, -, 1);
  }
}

int testline(coord start, coord end)
{ short xi, yi, dx, dy, d=0, i;

  dx=end.x-start.x;
  dy=end.y-start.y;

  xi=yi=1;

  if(dx<0) xi=-xi, dx=-dx;
  if(dy<0) yi=-yi, dy=-dy;

  if(dx>dy)
    for(i=0; i<dx; i++)
    { if(!viewable(start)) return 0;
      if(d+=dy>dx) d-=dx, start.y+=yi;
      start.x+=xi;      
    }
  else
    for(i=0; i<dy; i++)
    { if(!viewable(start)) return 0;
      if(d+=dx>dy) d-=dy, start.x+=xi;
      start.y+=yi;      
    }

  return 1;
}
